@@ -39,10 +39,11 @@ from urllib.error import HTTPError, URLError
 from urllib.parse import quote_plus
 from urllib.request import Request, urlopen
 
-from .common_utils import encode_int, string_to_bool
+from .common_utils import best_effort_deep_merge, coalesce, encode_int, string_to_bool
 from .config_resolver import ConfigResolver
 from .proc_wrapper_params import ProcWrapperParams, ProcWrapperParamValidationErrors
 from .runtime_metadata import (
+    CommonConfiguration,
     DefaultRuntimeMetadataFetcher,
     RuntimeMetadata,
     RuntimeMetadataFetcher,
@@ -113,6 +114,14 @@ class ProcWrapper:
 
     _STATUS_UPDATE_KEY_LAST_APP_HEARTBEAT_AT = "last_app_heartbeat_at"
 
+    _COPIED_RUNTIME_METADATA_PROPERTY_NAMES = [
+        "execution_method_type",
+        "infrastructure_type",
+        "allocated_cpu_units",
+        "allocated_memory_mb",
+        "ip_v4_addresses",
+    ]
+
     def __init__(
         self,
         params: Optional[ProcWrapperParams] = None,
@@ -121,9 +130,16 @@ class ProcWrapper:
         env_override: Optional[Mapping[str, Any]] = None,
         override_params_from_env: bool = True,
         runtime_context: Optional[Any] = None,
+        input_value: Optional[Any] = None,
+        override_params_from_input: bool = True,
+        override_params_from_config: Optional[bool] = None,
     ) -> None:
         _logger.info("Creating ProcWrapper instance ...")
 
+        self.input_value = input_value
+        self.runtime_context = runtime_context
+        self.override_params_from_env = override_params_from_env
+
         if env_override:
             self.env = dict(env_override)
         else:
@@ -162,8 +178,6 @@ class ProcWrapper:
             runtime_metadata_fetcher or DefaultRuntimeMetadataFetcher()
         )
 
-        self.runtime_context = runtime_context
-
         runtime_metadata: Optional[RuntimeMetadata] = None
         try:
             runtime_metadata = self.runtime_metadata_fetcher.fetch(
@@ -196,8 +210,6 @@ class ProcWrapper:
             want_env=True, want_config=self.params.embedded_mode
         )
 
-        self.config_resolver.fetch_and_resolve_env()
-
         if override_params_from_env:
             self.params.override_proc_wrapper_params_from_env(
                 env=self.resolved_env,
@@ -205,6 +217,26 @@ class ProcWrapper:
                 runtime_metadata=runtime_metadata,
             )
 
+        self.override_params_from_config = coalesce(
+            override_params_from_config, self.params.embedded_mode
+        )
+
+        if self.override_params_from_config:
+            env_override = self.params.override_proc_wrapper_params_from_config(
+                config=self.resolved_config, mutable_only=False
+            )
+
+            if env_override:
+                self.resolved_env.update(env_override)
+
+        if override_params_from_input:
+            env_override = self.params.override_proc_wrapper_params_from_input(
+                input=input_value
+            )
+
+            if env_override:
+                self.resolved_env.update(env_override)
+
         self.in_pytest = string_to_bool(os.environ.get("IN_PYTEST")) or False
 
         # Now we have enough info to try to send errors if problems happen below.
@@ -316,6 +348,62 @@ class ProcWrapper:
             last_app_heartbeat_at=last_app_heartbeat_at,
         )
 
+    def _transfer_runtime_metadata(
+        self,
+        dest: Dict[str, Any],
+        runtime_metadata: Optional[RuntimeMetadata],
+        override_props: Optional[Dict[str, Any]],
+        for_task: bool,
+    ) -> Dict[str, Any]:
+        em_details: Optional[Dict[str, Any]] = None
+        infra_settings: Optional[Dict[str, Any]] = None
+        rm_config: Optional[CommonConfiguration] = None
+
+        em_details_prop_name = (
+            "execution_method_capability_details"
+            if for_task
+            else "execution_method_details"
+        )
+
+        if runtime_metadata:
+            rm_config = runtime_metadata.task_configuration
+            infra_settings = rm_config.infrastructure_settings
+            em_details = rm_config.execution_method_capability_details
+
+            if not for_task:
+                rm_config = runtime_metadata.task_execution_configuration
+                em_details = (
+                    runtime_metadata.task_execution_configuration.execution_method_details
+                )
+
+        if override_props:
+            em_details = best_effort_deep_merge(
+                em_details, override_props.get(em_details_prop_name)
+            )
+
+            infra_settings = best_effort_deep_merge(
+                infra_settings, override_props.get("infrastructure_settings")
+            )
+
+        if em_details:
+            dest[em_details_prop_name] = em_details
+
+        if infra_settings:
+            dest["infrastructure_settings"] = infra_settings
+
+        for p in self._COPIED_RUNTIME_METADATA_PROPERTY_NAMES:
+            x = None
+            if rm_config:
+                x = getattr(rm_config, p)
+
+            if override_props:
+                x = coalesce(override_props.get(p), x)
+
+            if x is not None:
+                dest[p] = x
+
+        return dest
+
     def _create_or_update_task_execution(self) -> None:
         """
         Make a request to the API server to create a Task Execution
@@ -326,6 +414,7 @@ class ProcWrapper:
         if self.offline_mode:
             return
 
+        # TODO use runtime metadata
         if self.params.send_hostname and (self.hostname is None):
             try:
                 self.hostname = socket.gethostname()
@@ -419,6 +508,8 @@ class ProcWrapper:
             if stop_reason is not None:
                 body["stop_reason"] = stop_reason
 
+            rm = runtime_metadata if self.params.send_runtime_metadata else None
+
             if self.task_execution_uuid:
                 # Manually started
                 url += self.task_execution_uuid + "/"
@@ -462,21 +553,12 @@ class ProcWrapper:
 
                 task_dict["run_environment"] = run_env_dict
 
-                task_emc: Optional[Dict[str, Any]] = None
-                if self.params.send_runtime_metadata and runtime_metadata:
-                    task_emc = runtime_metadata.execution_method_capability
-
-                if self.params.auto_create_task_props:
-                    override_emc = self.params.auto_create_task_props.get(
-                        "execution_method_capability"
-                    )
-                    if override_emc:
-                        task_emc = task_emc or {}
-                        task_emc.update(override_emc)
-
-                task_dict["execution_method_capability"] = task_emc or {
-                    "type": "Unknown"
-                }
+                self._transfer_runtime_metadata(
+                    dest=task_dict,
+                    runtime_metadata=rm,
+                    override_props=self.params.auto_create_task_props,
+                    for_task=True,
+                )
 
                 body["task"] = task_dict
                 body["max_conflicting_age_seconds"] = self.params.max_conflicting_age
@@ -490,16 +572,16 @@ class ProcWrapper:
             if self.params.task_instance_metadata:
                 body["other_instance_metadata"] = self.params.task_instance_metadata
 
-            execution_method: Optional[Dict[str, Any]] = None
-            if self.params.send_runtime_metadata and runtime_metadata:
-                execution_method = runtime_metadata.execution_method
-
-            if self.params.execution_method_props:
-                execution_method = execution_method or {}
-                execution_method.update(self.params.execution_method_props)
-
-            if execution_method:
-                body["execution_method"] = execution_method
+            self._transfer_runtime_metadata(
+                dest=body,
+                runtime_metadata=rm,
+                override_props={
+                    # TODO
+                    # "execution_method_type": self.params.execution_method_type,
+                    "execution_method_details": self.params.execution_method_props
+                },
+                for_task=False,
+            )
 
             data = json.dumps(body).encode("utf-8")
 
@@ -666,7 +748,10 @@ class ProcWrapper:
 
         self._reload_params()
         self.log_configuration(initial=True)
-        self._setup_task_execution()
+
+        if not self._setup_task_execution():
+            self._exit_or_raise(self._EXIT_CODE_GENERIC_ERROR)
+            return
 
         rv = None
         success = False
@@ -730,17 +815,28 @@ class ProcWrapper:
             self.resolved_config,
             self.failed_config_props,
         ) = self.config_resolver.fetch_and_resolve_env_and_config(
-            want_env=True, want_config=self.params.embedded_mode
+            want_env=self.override_params_from_env,
+            want_config=self.override_params_from_config,
         )
 
         runtime_metadata = self.runtime_metadata_fetcher.fetch(
             env=self.resolved_env, context=self.runtime_context
         )
 
-        # In case API key(s) change
-        self.params.override_proc_wrapper_params_from_env(
-            env=self.resolved_env, mutable_only=True, runtime_metadata=runtime_metadata
-        )
+        if self.override_params_from_env:
+            self.params.override_proc_wrapper_params_from_env(
+                env=self.resolved_env,
+                mutable_only=True,
+                runtime_metadata=runtime_metadata,
+            )
+
+        if self.override_params_from_config:
+            env_override = self.params.override_proc_wrapper_params_from_config(
+                config=self.resolved_config, mutable_only=True
+            )
+
+            if env_override:
+                self.resolved_env.update(env_override)
 
         self.param_errors = self.params.validation_errors(
             runtime_metadata=runtime_metadata
@@ -1338,7 +1434,9 @@ class ProcWrapper:
         is_task_execution_creation_request: bool = False,
         is_final_update: bool = False,
     ) -> Optional[RawIOBase]:
-        _logger.debug(f"Sending {req.method} request to {req.full_url} ....")
+        _logger.debug(
+            f"Sending {req.method} request with body '{str(req.data)}' to {req.full_url} ...."
+        )
 
         self._refresh_api_server_retries_exhausted()
 
@@ -1441,9 +1539,8 @@ class ProcWrapper:
                 else:
                     self.last_api_request_failed_at = time.time()
 
-                    error_message = f"Got error response code = {status_code}"
+                    error_message = f"Got error response code = {status_code}, body = '{response_body}'"
                     _logger.error(error_message)
-
                     self._report_error(error_message, api_request_data)
 
                     if self.params.prevent_offline_execution:
@@ -1524,7 +1621,11 @@ class ProcWrapper:
         format = "PROC_WRAPPER: "
 
         if self.task_execution_uuid:
-            format += f"[{self.task_execution_uuid}] "
+            uuid_signature = self.task_execution_uuid
+            if len(uuid_signature) > 8:
+                uuid_signature = self.task_execution_uuid[0:8]
+
+            format += f"[{uuid_signature}] "
 
         if self.params.include_timestamps_in_log:
             format += "%(asctime)s "
